/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. See
 * http://www.eu-egee.org/partners/ for details on the copyright holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package org.glite.security.delegation.service;

import java.io.IOException;
import java.io.InputStream;
import java.io.BufferedInputStream;
import java.rmi.RemoteException;
import java.util.Calendar;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.glite.security.delegation.impl.GliteDelegation;

/**
 * Stub implementation for a SOAP frontend of the gLite Delegation Interface
 * (v1.1).
 * 
 * Makes calls to the actual logic implementation class.
 * 
 * @see org.glite.security.delegation.impl.GliteDelegation
 * 
 * @author Ricardo Rocha <Ricardo.Rocha@cern.ch>
 * @author Akos Frohner <Akos.Frohner@cern.ch>
 * @version $Id: DelegationSoapBindingImpl.java,v 1.9 2010/02/17 14:26:39 hahkala Exp $
 * 
 * Release: $Name: emi-delegation-service-java_R_1_5_1 $
 * 
 */
public class DelegationSoapBindingImpl implements Delegation {

    // Class logger
    private static Logger logger = Logger.getLogger(GliteDelegation.class);

    // Logic implementation class
    private GliteDelegation gliteDlg = null;

    /**
     * The name of the version file for the services.
     */
    static protected final String VERSION_PROPERTIES_FILE = "delegation.version.properties";

    /**
     * The singleton object.
     */
    static protected Properties props = null;

    /**
     * Initializes the class implementing the delegation logic.
     */
    public DelegationSoapBindingImpl() {
        try {
            gliteDlg = new GliteDelegation();
        } catch (IOException e) {
            logger.error("Failed to initialize the delegation service (failure while loading "
                            + "DLGEE properties config file).");
        }

        // for version and service metadata
        InitProperties();

        // Sanity check: see if the module version is known
        if (getProperty("module.version") == null) {
            logger.fatal("*** FATAL CONFIGURATION ERROR *** The module.version is unknown ***");
        }
    }

    public String getVersion() throws RemoteException, DelegationException {
        try {
            return getProperty("module.version");
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    public String getInterfaceVersion() throws RemoteException, DelegationException {
        try {
            return getProperty("interface.version");
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    public String getServiceMetadata(String key) throws RemoteException, DelegationException {
        try {
            return getProperty(key);
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getProxyReq(java.lang.String)
     */
    public java.lang.String getProxyReq(java.lang.String in0) throws DelegationException {

        String proxyReq = null;

        try {
            proxyReq = gliteDlg.getProxyReq(in0);
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }

        if (proxyReq != null) {
            return proxyReq;
        }

        return null;
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#putProxy(java.lang.String,
     *      java.lang.String)
     */
    public void putProxy(java.lang.String in0, java.lang.String in1) throws DelegationException {

        try {
            gliteDlg.putProxy(in0, in1);
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getNewProxyReq()
     */
    public NewProxyReq getNewProxyReq() throws RemoteException, DelegationException {

        try {
            org.glite.security.delegation.NewProxyReq implProxyReq = gliteDlg.getNewProxyReq(null);
            NewProxyReq proxyReq = new NewProxyReq();
            proxyReq.setDelegationID(implProxyReq.getDelegationID());
            proxyReq.setProxyRequest(implProxyReq.getProxyRequest());
            return proxyReq;
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#renewProxyReq(java.lang.String)
     */
    public String renewProxyReq(String dlgID) throws RemoteException, DelegationException {
        try {
            return gliteDlg.renewProxyReq(dlgID);
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#getTerminationTime(java.lang.String)
     */
    public Calendar getTerminationTime(String dlgID) throws RemoteException,
                    DelegationException {

        try {
            return gliteDlg.getTerminationTime(dlgID);
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    /**
     * @see org.glite.security.delegation.service.Delegation#destroy(java.lang.String)
     */
    public void destroy(String dlgID) throws RemoteException, DelegationException {

        try {
            gliteDlg.destroy(dlgID);
        } catch (org.glite.security.delegation.DelegationException de) {
            throw new DelegationException(de.getMsg());
        } catch (RuntimeException rt) {
            logger.error("Caught RuntimeException: ", rt);
            throw new DelegationException("Internal server error");
        }
    }

    static protected void InitProperties() {
        // return if the singleton is already initialized
        if (props != null) {
            return;
        }

        // singleton initialization
        props = new Properties();

        try {
            InputStream inp = null;

            ClassLoader loader = DelegationSoapBindingImpl.class.getClassLoader();

            if (loader != null) {
                inp = loader.getResourceAsStream(VERSION_PROPERTIES_FILE);
            } else {
                inp = ClassLoader.getSystemResourceAsStream(VERSION_PROPERTIES_FILE);
            }

            props.load(new BufferedInputStream(inp));
            inp.close();

            logger.info("Configuration file '" + VERSION_PROPERTIES_FILE + "' loaded");
        } catch (IOException e) {
            logger.error("Error loading config file " + VERSION_PROPERTIES_FILE + ": " + e);
        }
    }

    /**
     * Returns the value of the given property, or <code>null</code> if
     * there is no such property.
     */
    static protected String getProperty(String key) {
        String value = props.getProperty(key);
        logger.debug("Property " + key + " = " + value);

        return value;
    }

}
