/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. See
 * http://www.eu-egee.org/partners/ for details on the copyright holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */

package org.glite.security.delegation;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.cert.X509Certificate;
import java.util.Calendar;

import javax.xml.rpc.ServiceException;

import junit.framework.TestCase;

import org.apache.axis.AxisFault;
import org.apache.axis.configuration.SimpleProvider;
import org.apache.log4j.Appender;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.Layout;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.PatternLayout;
import org.bouncycastle.jce.PKCS10CertificationRequest;
import org.bouncycastle.openssl.PEMReader;
import org.glite.security.delegation.service.Delegation;
import org.glite.security.delegation.service.DelegationException;
import org.glite.security.delegation.service.DelegationServiceLocator;
import org.glite.security.delegation.service.NewProxyReq;
import org.glite.security.trustmanager.axis.SSLConfigSender;
import org.glite.security.util.FileCertReader;
import org.glite.security.util.PrivateKeyReader;
import org.glite.security.util.proxy.ProxyCertificateGenerator;

/**
 * Functional tests for the gLite Delegation Interface (v1.1).
 *
 * @Author  Ricardo Rocha <Ricardo.Rocha@cern.ch>
 * @Author  Joni Hahkala <joni.hahkala@cern.ch>
 *
 */
public class BasicDelegationTestFunc extends TestCase {
    
    /**
     * Class logger.
     */
    private final static Logger s_log = Logger.getLogger(BasicDelegationTestFunc.class);

    /**
     * The default endpoint to run the tests against.
     */
//    private final String DLG_ENDPOINT = "https://pchip10.cern.ch:8443/glite-security-delegation-service-java/services/gridsite-delegation";
    private final String DLG_ENDPOINT = "https://localhost:8443/delegation-service-java/services/gridsite-delegation";

    /**
     * The default DLG ID to use.
     */
    private final String DLG_ID = "my-test-request-id";

    /**
     * The "name" of one set of client credentials to use in tests.
     */
    private final String DEFAULT_CLIENT_1 = "/home/hahkala/workspace/Trustmanager/test/certs/trusted-certs/trusted_client.proxy.grid_proxy";

    /**
     * The "name" of another set of client credentials to use in tests.
     */
    private final String DEFAULT_CLIENT_2 = "/home/hahkala/workspace/Trustmanager/test/certs/trusted-certs/trusted_server.proxy.grid_proxy";

    /**
     * Default lifetime to give to the proxy.
     */
    private final int PROXY_LIFETIME = 12;

    /**
     * Actually endpoint holder being used for the tests.
     */
    private String m_dlgEndpoint = null;

    /**
     * Holder of the stub used to talk to the service.
     */
    private Delegation m_stubDlg = null;

    /**
     * Holder of the DLGOR config properties.
     */
    private GrDProxyDlgorOptions m_dlgorOpt = null;

    /**
     * Holder of the proxy generator utility.
     */
//    private GrDProxyGenerator m_proxyGenerator = null;

    /**
     * Location of the client credentials used in tests.
     */
    private String m_certsDir = null;
    
    /**
     * the Certificate and proxy reader class.
     */
    private FileCertReader m_reader = null;
    
    /**
     * Initializes all the required test utilities (proxy generators,
     * stubs, config holders, ...).
     *
     * @throws MalformedURLException The endpoint to use in the tests is invalid
     * @throws ServiceException Failed to create a connection to the test endpoint
     * @throws Exception Any other error
     */
    public BasicDelegationTestFunc() throws MalformedURLException, ServiceException, IOException {
        s_log.debug("Entered BasicDelegationTestFunc.");
        // Get endpoint to use
        m_dlgEndpoint = System.getProperty("dlgEndpoint");
        if (m_dlgEndpoint == null) {
            m_dlgEndpoint = DLG_ENDPOINT;
        }

        // Create stub for Delegation interface
        s_log.debug("Delegation Endpoint: " + m_dlgEndpoint);
        SimpleProvider provider = SSLConfigSender.getTransportProvider(System.getProperties());
        DelegationServiceLocator locator = new DelegationServiceLocator(provider);
        m_stubDlg = locator.getGridsiteDelegation(new URL(m_dlgEndpoint));
        s_log.debug("Delegation stub created successfully.");

        // Create the DLGOR properties object
        File file = new File(".");
//        System.out.println(file.getAbsolutePath());
        m_dlgorOpt = new GrDProxyDlgorOptions("config/dlgor.properties"/*GrDPX509Util.getDlgorPropertyFile()*/);
        /**s_log.debug("DLGOR properties loaded.");
        s_log.debug("User Cert File: " + m_dlgorOpt.getDlgorCertFile());
        s_log.debug("User Private Key File: " + m_dlgorOpt.getDlgorKeyFile());
        s_log.debug("User Proxy: " + m_dlgorOpt.getDlgorProxyFile());
        s_log.debug("User Password: " + m_dlgorOpt.getDlgorPass());*/

        try{
            m_reader = new FileCertReader();
        }catch (Exception e){
            s_log.error("Internal error while generating FileCertReader: " + e.getMessage());
        }
        
        // Create a Proxy Generator
//        m_proxyGenerator = new GrDProxyGenerator();
//        m_proxyGenerator.setLifetime(PROXY_LIFETIME);

        // Get location of certificates
        m_certsDir = System.getProperty("stageCertDir");
        
    }

    protected void setUp() {

    }

    protected void tearDown() throws RemoteException {

    }

    public void testProxyDelegation() {
        s_log.debug("Entered testProxyDelegation.");

        destroy(DEFAULT_CLIENT_1, DLG_ID);
        
        setActiveUser(DEFAULT_CLIENT_1);
        
        String certRequest = null;
        try {
            s_log.debug("Calling getProxyReq.");
            certRequest = m_stubDlg.getProxyReq(DLG_ID);
            s_log.debug("Certificate Request:\n" + certRequest);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        try {
            // Generate the proxy certificate
            String proxyCert = this.getProxy(certRequest);
            s_log.debug("Generated proxy certificate: " + proxyCert);

            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(DLG_ID, proxyCert);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        s_log.debug("Certificate Request: \n" + certRequest);
    }

    public void testProxyDelegationWithProxy() {
        s_log.debug("Entered testProxyDelegationWithProxy.");

        setActiveUser(DEFAULT_CLIENT_2);

        destroy(DEFAULT_CLIENT_2, DLG_ID);

        String certRequest = null;
        try {
            s_log.debug("Calling getProxyReq.");
            certRequest = m_stubDlg.getProxyReq(DLG_ID);
            s_log.debug("Certificate Request:\n" + certRequest);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        try {
            // Generate the proxy certificate
            String proxyCert = this.getProxy(certRequest);
            s_log.debug("Generated proxy certificate: " + proxyCert);

            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(DLG_ID, proxyCert);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        s_log.debug("Certificate Request: \n" + certRequest);
    }

    public void testProxyDelegationNullDlgID() {
        s_log.debug("Entered testProxyDelegationNullDlgID.");

        setActiveUser(DEFAULT_CLIENT_1);

        destroy(DEFAULT_CLIENT_1, null);

        // Make the request
        String certRequest = null;
        try {
            s_log.debug("Calling getProxyReq.");
            certRequest = m_stubDlg.getProxyReq(null);
            s_log.debug("Certificate Request:\n" + certRequest);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        try {
            // Generate the proxy certificate
            String proxyCert = this.getProxy(certRequest);
            s_log.debug("Generated proxy certificate: " + proxyCert);

            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(null, proxyCert);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        s_log.debug("Certificate Request: \n" + certRequest);
    }

    public void testPutProxyWithNullProxy() throws Exception {
        s_log.debug("Entered testPutProxyWithNullProxy.");

        setActiveUser(DEFAULT_CLIENT_1);

        destroy(DEFAULT_CLIENT_1, null);

        // Make the request
        String certRequest = null;
        s_log.debug("Calling getProxyReq.");
        certRequest = m_stubDlg.getProxyReq(null);
        s_log.debug("Certificate Request:\n" + certRequest);

        try {
            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(null, null);
        } catch (DelegationException e) {
            // Success
            s_log.debug("Caught exception as expected. Exception was: " + e.getMessage());
            return;
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        assertTrue("Should have caught a DelegationException.", false);
    }

    public void testProxyDelegationWithWrongProxy() {
        Logger LOGGERcrl = Logger.getLogger("org.glite.security.util.FileCRLChecker");
        Logger LOGGERfr = Logger.getLogger("org.glite.security.util.FileCertReader");
        Logger LOGGERRoot = Logger.getLogger("org.glite.security");
        Layout lay = new PatternLayout("%d{ISO8601} %-5p [%t] %c{2}%l %x - %m%n");
        Appender appender = new ConsoleAppender(lay);
        LOGGERRoot.addAppender(appender);
        LOGGERRoot.setLevel(Level.INFO);
        LOGGERfr.setLevel(Level.OFF);
        LOGGERcrl.setLevel(Level.OFF);
       s_log.debug("Entered testProxyDelegationWithWrongProxy.");

        setActiveUser(DEFAULT_CLIENT_1);

        destroy(DEFAULT_CLIENT_1, null);

        String certRequest = null;
        try {
            s_log.debug("Calling getProxyReq.");
            certRequest = m_stubDlg.getProxyReq(null);
            s_log.debug("Certificate Request:\n" + certRequest);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        try {
            // Generate the proxy certificate
            String proxyCert = this.getProxy(certRequest);
            s_log.debug("Generated proxy certificate: " + proxyCert);

            // Set a different user for the putProxy call
            setActiveUser(DEFAULT_CLIENT_2);

            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(null, proxyCert);
        } catch (DelegationException e) {
            // Success
            System.out.println("got expected exception: ");
            e.printStackTrace(System.out);
            s_log.debug("Caught exception as expected. Exception was: " + e.getMessage());
            return;
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        assertTrue("Should have caught a DelegationException.", false);
    }

    public void testGetTerminationTime() {
        s_log.debug("Entered testGetTerminationTimeWithNullID.");

        // Delegate
        String dlgID = delegateCredential(DEFAULT_CLIENT_1);

        // Test it first passing null as DLG ID (will take it from client info)
        try {
            s_log.debug("Calling getTerminationTime with NULL id.");
            Calendar cal = m_stubDlg.getTerminationTime(null);
            // TODO: check that the time is correct
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        // Test it again passing the DLG ID explicitly
        try {
            s_log.debug("Calling getTerminationTime with NULL id.");
            Calendar cal = m_stubDlg.getTerminationTime(dlgID);
            // TODO: check that the time is correct
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }
    }

    public void testDestroyDelegationWithNullID() {
        s_log.debug("Entered testDestroyDelegationWithNullID.");

        // Delegate
        String dlgID = delegateCredential(DEFAULT_CLIENT_1);

        try {
            s_log.debug("Calling testDestroyDelegationWithNullID with NULL id.");
            m_stubDlg.destroy(null);
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }
    }

    public void testDestroyDelegationWithGivenID() {
        s_log.debug("Entered testDestroyDelegationWithGivenID.");

        // Delegate
        String dlgID = delegateCredential(DEFAULT_CLIENT_1);

        try {
            s_log.debug("Calling testDestroyDelegationWithGivenID.");
            m_stubDlg.destroy(dlgID);
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }
    }

    public void testDestroyDelegationWithWrongClientProxy() {
        s_log.debug("Entered testDestroyDelegationWithWrongClientProxy.");

        // Delegate
        String dlgID = delegateCredential(DEFAULT_CLIENT_1);

        setActiveUser(DEFAULT_CLIENT_2);
        try {
            s_log.debug("Calling testDestroyDelegationWithWrongClientProxy.");
            m_stubDlg.destroy(dlgID);
        } catch (DelegationException e) {
            // Nice
            System.out.println("got expected exception: ");
            e.printStackTrace(System.out);
            return;
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        assertTrue("Should have received an exception.", false);
    }

    public void testGetTerminationTimeWithWrongClientProxy() {
        s_log.debug("Entered testGetTerminationTimeWithWrongClientProxy.");

        // Delegate
        String dlgID = delegateCredential(DEFAULT_CLIENT_1);

        setActiveUser(DEFAULT_CLIENT_2);
        try {
            s_log.debug("Calling testGetTerminationTimeWithWrongClientProxy.");
            m_stubDlg.getTerminationTime(dlgID);
        } catch (DelegationException e) {
            // Nice
            System.out.println("got expected exception: ");
            e.printStackTrace(System.out);
            return;
        } catch (RemoteException e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        assertTrue("Should have received an exception.", false);
    }

    /**
     * Creates a proxy from the given certificate request and returns it.
     * 
     * @param certificateRequest The certificate request for proxy creation
     * @return The proxy certificate
     * @throws Exception Failed to create the proxy certificate from the certificate request
     */
    private String getProxy(String certificateRequest) throws Exception {

        // Get location of the user certificate files
        String proxyFileLocation = System.getProperties().getProperty("gridProxyFile");
        String certFileLocation = System.getProperties().getProperty("sslCertFile");
        String keyFileLocation = System.getProperties().getProperty("sslKey");
        
        // Declare holders for the certificate files
/*        byte[] certFile = null;
        byte[] keyFile = null;
        byte[] proxyFile = null;
        byte[] certRequest = certificateRequest.getBytes();
*/
        // And the final signed proxy certificate holder
        String signedProxyCert = null;
        
        // change the request into a proper instance of a class
        PEMReader pemReader = new PEMReader(new StringReader(certificateRequest));
        PKCS10CertificationRequest req = (PKCS10CertificationRequest)pemReader.readObject();

        // Load the certificate files
        try {
            if (proxyFileLocation != null) {
                // Load only the proxy file if it was set
                KeyStore store = m_reader.readProxy(new BufferedInputStream(new FileInputStream(proxyFileLocation)), "");
                X509Certificate[] certs = (X509Certificate[])store.getCertificateChain("host");
                PrivateKey key = (PrivateKey) store.getKey("host", "".toCharArray());
                // Create the proxy certificate from the request
                ProxyCertificateGenerator generator = new ProxyCertificateGenerator(certs, req);
                generator.generate(key);
                signedProxyCert = generator.getCertChainAsPEM();
                
            } else if (certFileLocation != null && keyFileLocation != null) {
                
                // Load certificate and private key files
                X509Certificate cert = (X509Certificate)m_reader.readCerts(certFileLocation).firstElement();
                PrivateKey key = PrivateKeyReader.read(new BufferedInputStream(new FileInputStream(keyFileLocation)), m_dlgorOpt.getDlgorPass());
                ProxyCertificateGenerator generator = new ProxyCertificateGenerator(cert, req);
                generator.generate(key);
                signedProxyCert = generator.getCertChainAsPEM();
            } else {
                s_log.error("Failed to find proxy or cert/key files. Check DLGOR config file.");
                assertTrue("Failed to find proxy or cert/key files. Check DLGOR config file.",
                                false);
            }
        } catch (FileNotFoundException e) {
            s_log.error("Failed while accessing user proxy/cert/key files: " + e.getMessage());
            assertTrue("Failed while accessing user proxy/cert/key files: " + e.getMessage(), false);
        }

        if (signedProxyCert == null) {
            s_log.error("Failed to generate a proxy.");
            assertTrue("Failed to generate a proxy.", false);
        }

        return signedProxyCert;
    }

    /**
     * Sets as current active credentials the ones pointed by the argument given.
     * 
     * @param userID The ID of the set of credentials to be used.
     */
    private void setActiveUser(String userID) {
        if (userID.startsWith("/")) {
            System.getProperties().remove("sslCertFile");
            System.getProperties().remove("sslKey");
            System.getProperties().remove("sslKeyPasswd");
            System.getProperties().setProperty("gridProxyFile", userID);
        } else {
            System.getProperties().setProperty("sslCertFile",
                            m_certsDir + "/home/usercert" + userID + ".pem");
            System.getProperties().setProperty("sslKey",
                            m_certsDir + "/home/userkey" + userID + ".pem");
            System.getProperties().setProperty("sslKeyPasswd", "changeit");
            System.getProperties().remove("gridProxyFile");
        }
        try {
            System.getProperties().list(System.out);
            System.out.println("-----------------------" + System.getProperty("gridProxyFile"));
            SimpleProvider provider = SSLConfigSender.getTransportProvider(System.getProperties());
            DelegationServiceLocator locator = new DelegationServiceLocator(provider);
            m_stubDlg = locator.getGridsiteDelegation(new URL(m_dlgEndpoint));
        } catch (AxisFault e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (MalformedURLException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        } catch (ServiceException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * Performs delegation using the set of credentials specified as argument.
     * 
     * Delegation is performed using the getNewProxyReq method.
     * 
     * @param activeUser The set of credentials to be used for delegation
     * @return The delegation ID assigned to the new set of credentials.
     */
    private String delegateCredential(String activeUser) {
        setActiveUser(DEFAULT_CLIENT_1);

        // Destroy the credential if it exists
        try {
            m_stubDlg.destroy(null);
        } catch (Exception e) {
            // catch this one
            System.out.println("got expected exception: ");
            e.printStackTrace(System.out);
        }

        // Delegate a credential
        NewProxyReq proxyReq = null;
        try {
            s_log.debug("Calling getNewProxyReq.");
            proxyReq = m_stubDlg.getNewProxyReq();
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        try {
            // Generate the proxy certificate
            String proxyCert = this.getProxy(proxyReq.getProxyRequest());
            s_log.debug("Generated proxy certificate: " + proxyCert);

            // Put the proxy
            s_log.debug("Calling putProxy.");
            m_stubDlg.putProxy(proxyReq.getDelegationID(), proxyCert);
        } catch (Exception e) {
            s_log.fatal("Got exception from service: " + e.getMessage());
            e.printStackTrace();
            assertTrue("Got exception from service: " + e.getMessage(), false);
        }

        return proxyReq.getDelegationID();
    }

    /**
     * Destroys the specified credential in case it exists.
     * 
     * Does not give any error report back (eats the exception if it occurs).
     *
     * @param activeUser The set of credentials to be used for the call
     * @param dlgID The delegation ID of the credentials to be destroyed
     */
    private void destroy(String activeUser, String dlgID) {
        
        setActiveUser(activeUser);

        try {
            m_stubDlg.destroy(dlgID);
        } catch (DelegationException e) {
            // catch this one
        } catch (RemoteException e) {
            s_log.debug("Got exception while setting up test. Exception: " + e);
            assertTrue("Got exception while setting up test. Exception: " + e, false);
        }
    }

}
